#include "mxcl_sys.h"



#ifndef MAX
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#endif

#ifndef True
#define True 1
#endif
#ifndef False
#define False 0
#endif




/* application constants */

#define MAX_CHAR 24
#define MAX_NAME_CHAR  16
#define MAX_SIGN_CHAR   8
/*#define MAX_DISP_CHAR  10*/
#define MAX_DISP_CHAR  20
#define MAX_LARGE_FONT 14

#define MAX_VEC  4

#define NUM_INF   1
//#define NUM_INT   1
#define NUM_FLOAT 2
#define NUM_BIN   3
#define NUM_OCT   4
#define NUM_HEX   5
#define NUM_TIME  6
#define NUM_DATE  7

#define USYS_NONE  1
#define UNIT_NONE  1
#define USYS_TIME  4
#define UNIT_HMS   1
#define UNIT_SEC   2

#define ST_BF 1
#define ST_OP 2

#define RD_NONE 0
#define RD_DISP 1
#define RD_REAL 2
#define RD_BELW 1
#define RD_SGNF 2
#define RD_DOWN 1
#define RD_OFF  2
#define RD_UP   3

#define F_PRE 1
#define F_LOCAL 2

#define ARG_SD 99

#define PREF_DMIN 0.00000001
#define PREF_DMAX  100000000

#define STR_INF  "Inf"
#define STR_MINF "-Inf"
#define STR_NaN  "NaN"

#define PI  3.14159265
#define LNE 2.71828183



/* type definition */

typedef struct {
    UInt16 op;
    UInt16 st;
    Boolean mm;
    Boolean fx;
} OperationStatus;

extern OperationStatus op_st;

typedef struct {
    Boolean vl;
    Boolean op;
    Boolean ns;
    Boolean pb;
    Boolean un;
    Boolean pp;
    Boolean mk;
    Boolean fr;
} UpdateStatus;

extern UpdateStatus update;

typedef struct {
    Boolean fix;
    Boolean up;
    UInt16 usys;
    UInt16 unit;
    UInt16 nsys;
    UInt16 nstr;
    UInt16 nvec;
    Char str[MAX_CHAR];
    FlpCompDouble v[MAX_VEC];
} ValueStruct;

typedef struct {
    UInt16 usys;
    UInt16 unit;
    double rate;
    Char sign[MAX_SIGN_CHAR];
    Char name[MAX_NAME_CHAR];
} UnitStruct;

typedef struct {
    UInt8 disp;
    UInt8 type;
    UInt8 cond;
    Int8  b_dgt;
    Int8  s_dgt;
} RoundType;


typedef double ConstFunc(void);
typedef void   ConstLocalFunc(ValueStruct*);
typedef double SingleArgFunc(double);
typedef void   SingleArgLocalFunc(ValueStruct*);
typedef double DoubleArgFunc(double, double);
typedef void   DoubleArgLocalFunc(ValueStruct*, ValueStruct*);


typedef struct {
    UInt16 id;
    UInt8 args;
    UInt8 type;
    Boolean mathlib;
    void* func;
} OperatorList;


/* functions declaration */

#include "mxcl_sys_proto.h"


/* mxcl.c */

extern void QuitApp(void);


/* mxcl_handle.c */

extern void InitializeStatus(void);
extern void InitializeVariables(void);
extern void GetStatus(void);
extern void SaveStatus(void);
extern void OpenFormMain(FormPtr);
extern Boolean FormMainHandleEventPalm(EventPtr);

/* mxcl_form_unit.c */

extern void OpenFormUnitList(FormPtr);
extern void CloseFormUnitList(void);



/* mxcl_db.c */

extern void GetHMS(UInt16*, UInt16*, UInt16*);
extern void InitializeUnitStruct(UnitStruct*);
extern double GetUnitRate(UInt16, UInt16);
extern void GetUnitSign(UInt16, UInt16, Char*);
extern void GetUnitName(UInt16, UInt16, Char*);
extern void GetUnitStruct(UInt16, UInt16, UnitStruct*);
extern UInt16 GetUnitSystemNum(UInt16);
extern void AddUnitStruct(UnitStruct*);
extern void SaveUnitStruct(UnitStruct*);
extern Boolean DeleteUnitStruct(UnitStruct*);
extern Boolean DownUnitStruct(UnitStruct*);
extern void AddUnitSystemStruct(UnitStruct*);
extern Boolean DeleteUnitSystemStruct(UnitStruct*);
extern Boolean DownUnitSystemStruct(UnitStruct*);
extern void InitializeUnitDB(void);
extern Err OpenUnitDB(Boolean);
extern void CloseUnitDB(void);




/* mxcl_util.c */

extern UInt16 PushBtnIDToFormID(UInt16);
extern UInt16 FormIDToPushBtnID(UInt16);
extern UInt16 PushBtnIDToNumsys(UInt16);
extern UInt16 NumsysToPushBtnID(UInt16);
extern WChar ButtonIDToChar(UInt16);
extern WChar MemBtnIDToChar(UInt16);
extern UInt16 CharToNumsys(Char, UInt16);
extern UInt16 NumSysToStr(Char*, UInt16);
extern Boolean IsAddChar(WChar);
extern Boolean IsOperatorDouble(WChar);
extern Boolean IsOperatorDoubleBtn(UInt16);
extern Boolean IsToBeThSeparated(ValueStruct*);
extern Boolean IsToBeBitSeparated(ValueStruct*);
extern Boolean ParseNoError(Char*, UInt16, UInt16);
extern void ChopZero(ValueStruct*, WChar);
extern void ThSeparateNumber(Char*);
extern void BitSeparateNumber(Char*);

extern void ValFloatToStr(RoundType, FlpCompDouble, Char*);
extern Boolean StrToVal(ValueStruct*);
extern void ValToStr(ValueStruct*);



/* mxcl_prefs.c */

extern UInt16 PrefCurrentForm(void);
extern void PrefSetCurrentForm(UInt16);
extern Char PrefDecimalSeparator(void);
extern Char PrefThousandSeparator(void);
extern void PrefDMinMax(double*, double*);
extern void PrefSetDMinMax(double, double);
extern RoundType PrefRoundType(void);
extern Boolean PrefZeroPadding(void);
extern Boolean PrefUBit(void);
extern void GetPrefs(void);
extern void SavePrefs(void);


/* mxcl_func.c */

extern void SetOperatorsFunc(void) SEC_FC;

extern void VecRoundType(RoundType*);
extern void RoundValue(RoundType, FlpCompDouble*);
extern void InitializeValue(ValueStruct*);
extern void CopyValue(ValueStruct*, ValueStruct*);
extern void SetConstValue(ValueStruct*, double);

extern void ConvertUnit(UInt16, UInt16, ValueStruct*);
extern void SetNumsys(UInt16, ValueStruct*);
extern void SetUnit(UInt16, UInt16, ValueStruct*) SEC_FC;

extern void OpenMathLib(void) SEC_FC;
extern void CloseMathLib(void) SEC_FC;
extern Boolean IsMathLibOpened(void) SEC_FC;

extern UInt16 SingleToDoubleOpeID(UInt16) SEC_FC;
extern OperatorList GetOperatorList(UInt16) SEC_FC;
extern void OperateSingleValue(OperatorList, ValueStruct*) SEC_FC;
extern void OperateTwoValues(UInt16, ValueStruct*, ValueStruct*) SEC_FC;
extern void OperateConstValue(OperatorList, ValueStruct*) SEC_FC;
