#include <stddef.h>
#include "mxcl.h"
#include "mxcl_id.h"

#define DB_DEFAULT

/*#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#enfif*/


DmOpenRef udb=0;

UInt16 usys_time=USYS_TIME;
UInt16 unit_hms=UNIT_HMS;
UInt16 unit_sec=UNIT_SEC;


/* local prototypes */

extern Boolean SeekHMSInUsys(UInt16);
extern void SeekHMS(void);
extern Int16 CompareUnit(UnitStruct*, UnitStruct*, Int16,
                         SortRecordInfoPtr, SortRecordInfoPtr,
                         MemHandle);
extern Err CreateUnitDB(void);



/* util for HMS */

void GetHMS(UInt16* us_t, UInt16* un_h, UInt16* un_s)
{
    *us_t=usys_time;
    *un_h=unit_hms;
    *un_s=unit_sec;
}

Boolean SeekHMSInUsys(UInt16 usys)
{
    Boolean ret=False;
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;
    UnitStruct ust={0, 0, 1.0, ""};

    ust.usys=usys;
    ust.unit=1;
    id=DmFindSortPosition(udb, &ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    id--;
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmQueryRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys>ust.usys){
            MemHandleUnlock(dh);
            break;
        }else if(StrCompare(dp->name, "HMS")==0){
            unit_hms=dp->unit;
            ret=True;
        }else if(dp->rate==1.0){
            unit_sec=dp->unit;
        }
        MemHandleUnlock(dh);
        id++;
    }
    return ret;
}

void SeekHMS()
{
    UInt16 usys=1;

    for(usys=1; usys<=GetUnitSystemNum(0); usys++){
        if(SeekHMSInUsys(usys)){
            usys_time=usys;
            break;
        }
    }
}


/* database operations */

void InitializeUnitStruct(UnitStruct* ust)
{
    ust->usys=0;
    ust->unit=0;
    ust->rate=1.0;
    ust->sign[0]='\0';
    ust->name[0]='\0';
}

Int16 CompareUnit(UnitStruct* rec1,
                  UnitStruct* rec2,
                  Int16 prm,
                  SortRecordInfoPtr info1,
                  SortRecordInfoPtr info2,
                  MemHandle appInfoH)
{
    Int16 ret;

    CALLBACK_PROLOGUE

    ret=rec1->usys-rec2->usys;
    if(ret==0){
        ret=rec1->unit-rec2->unit;
    }
    CALLBACK_EPILOGUE
    return ret;
}

double GetUnitRate(UInt16 usys, UInt16 unit)
{
    double ret=1.0;
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;
    UnitStruct ust={0, 0, 1.0, ""};

    ust.usys=usys;
    ust.unit=unit;
    id=DmFindSortPosition(udb, &ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    id--;
    dh=DmQueryRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys==usys && dp->unit==unit){
        ret=dp->rate;
    }
    MemHandleUnlock(dh);
    return ret;
}

void GetUnitSign(UInt16 usys, UInt16 unit, Char* str)
{
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;
    UnitStruct ust={0, 0, 1.0, ""};

    str[0]='\0';
    ust.usys=usys;
    ust.unit=unit;
    id=DmFindSortPosition(udb, &ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return;
    id--;
    dh=DmQueryRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys==usys && dp->unit==unit){
        StrCopy(str, dp->sign);
    }
    MemHandleUnlock(dh);
    return;
}

void GetUnitName(UInt16 usys, UInt16 unit, Char* str)
{
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;
    UnitStruct ust={0, 0, 1.0, ""};

    str[0]='\0';
    ust.usys=usys;
    ust.unit=unit;
    id=DmFindSortPosition(udb, &ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return;
    id--;
    dh=DmQueryRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys==usys && dp->unit==unit){
        StrCopy(str, dp->name);
    }
    MemHandleUnlock(dh);
    return;
}

void GetUnitStruct(UInt16 usys, UInt16 unit, UnitStruct* ust)
{
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;

    ust->usys=usys;
    ust->unit=unit;
    id=DmFindSortPosition(udb, ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return;
    id--;
    dh=DmQueryRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys==usys && dp->unit==unit){
        ust->rate=dp->rate;
        StrCopy(ust->sign, dp->sign);
        StrCopy(ust->name, dp->name);
    }
    MemHandleUnlock(dh);
    return;
}

UInt16 GetUnitSystemNum(UInt16 usys)
{
    Int16 i=0;
    UInt16 id, num;
    MemHandle dh;
    UnitStruct* dp;
    UnitStruct ust={0, 0, 1.0, ""};

    ust.usys=usys;
    id=DmFindSortPosition(udb, &ust, 0, (DmComparF*)CompareUnit, 0);
    num=DmNumRecordsInCategory(udb, dmAllCategories);
    while(id<num){
        dh=DmQueryRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys==usys){
            i++;
            MemHandleUnlock(dh);
        }else{
            MemHandleUnlock(dh);
            break;
        }
        id++;
    }
    return i;
}


/* unit struct operations */

void AddUnitStruct(UnitStruct* ust)
{
    UInt16 id, unit;
    MemHandle dh;
    UnitStruct* dp;

    ust->unit--;
    id=DmFindSortPosition(udb, ust, 0, (DmComparF*)CompareUnit, 0);
    ust->unit++;
    dh=DmNewRecord(udb, &id, sizeof(*ust));
    dp=(UnitStruct*)MemHandleLock(dh);
    DmWrite(dp, 0, ust, sizeof(*ust));
    MemHandleUnlock(dh);
    DmReleaseRecord(udb, id, True);

    id++;
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmGetRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys!=ust->usys){
            MemHandleUnlock(dh);
            DmReleaseRecord(udb, id, True);
            break;
        }
        if(dp->unit>=ust->unit){
            unit=dp->unit;
            unit++;
            DmWrite(dp, offsetof(UnitStruct, unit), &unit, sizeof(unit));
        }
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        id++;
    }
}

void SaveUnitStruct(UnitStruct* ust)
{
    UInt16 id;
    MemHandle dh;
    UnitStruct* dp;

    id=DmFindSortPosition(udb, ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return;
    id--;
    dh=DmGetRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys==ust->usys && dp->unit==ust->unit){
/*        dp->rate=ust->rate;
        StrCopy(dp->sign, ust->sign);
        StrCopy(dp->name, ust->name);*/
        DmWrite(dp, 0, ust, sizeof(*ust));
    }else{
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        dh=DmNewRecord(udb, &id, sizeof(*ust));
        dp=(UnitStruct*)MemHandleLock(dh);
        DmWrite(dp, 0, ust, sizeof(*ust));
    }
    MemHandleUnlock(dh);
    DmReleaseRecord(udb, id, True);
//    if(ust->usys==usys_time) SeekHMS(ust->usys);
}

Boolean DeleteUnitStruct(UnitStruct* ust)
{
    Boolean ret=False;
    UInt16 id, alert_id, unit;
    MemHandle dh;
    UnitStruct* dp;

    if(ust->usys==0){
        alert_id=ID_ALERT_DEL_US;
    }else{
        alert_id=ID_ALERT_DEL_UN;
    }
    id=DmFindSortPosition(udb, ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    id--;
    dh=DmQueryRecord(udb, id);
    dp=(UnitStruct*)MemHandleLock(dh);
    if(dp->usys!=ust->usys || dp->unit!=ust->unit
       || FrmCustomAlert(alert_id, dp->name, NULL, NULL)>0){
        MemHandleUnlock(dh);
        return ret;
    }
    MemHandleUnlock(dh);
    DmDeleteRecord(udb, id);
    DmMoveRecord(udb, id, DmNumRecords(udb));

    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmGetRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys!=ust->usys){
            MemHandleUnlock(dh);
            DmReleaseRecord(udb, id, True);
            break;
        }
        if(dp->unit>=ust->unit){
            unit=dp->unit;
            unit--;
            DmWrite(dp, offsetof(UnitStruct, unit), &unit, sizeof(unit));
        }
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        id++;
    }
    if(GetUnitSystemNum(ust->usys)==0){
        UnitStruct u;

        InitializeUnitStruct(&u);
        u.usys=ust->usys;
        u.unit=1;
        AddUnitStruct(&u);
    }
//    if(ust->usys==usys_time) SeekHMS(ust->usys);
    return True;
}

Boolean DownUnitStruct(UnitStruct* ust)
{
    Boolean ret=False;
    UInt16 id, unit;
    MemHandle dh1, dh2;
    UnitStruct* dp1;
    UnitStruct* dp2;

    id=DmFindSortPosition(udb, ust, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    if(id>=DmNumRecordsInCategory(udb, dmAllCategories)) return ret;
    id--;
    dh1=DmGetRecord(udb, id);
    dp1=(UnitStruct*)MemHandleLock(dh1);
    if(dp1->usys!=ust->usys || dp1->unit!=ust->unit){
        MemHandleUnlock(dh1);
        DmReleaseRecord(udb, id, True);
        return ret;
    }
    dh2=DmGetRecord(udb, id+1);
    dp2=(UnitStruct*)MemHandleLock(dh2);
    if(dp2->usys!=ust->usys || dp2->unit!=ust->unit+1){
        MemHandleUnlock(dh1);
        DmReleaseRecord(udb, id, True);
        MemHandleUnlock(dh2);
        DmReleaseRecord(udb, id+1, True);
        return ret;
    }
    unit=dp1->unit;
    unit++;
    DmWrite(dp1, offsetof(UnitStruct, unit), &unit, sizeof(unit));
    MemHandleUnlock(dh1);
    DmReleaseRecord(udb, id, True);
    unit=dp1->unit;
    unit--;
    DmWrite(dp2, offsetof(UnitStruct, unit), &unit, sizeof(unit));
    MemHandleUnlock(dh2);
    DmReleaseRecord(udb, id+1, True);
    DmMoveRecord(udb, id+1, id);
//    if(ust->usys==usys_time) SeekHMS(ust->usys);
    return True;
}

/* unit system operations */

void AddUnitSystemStruct(UnitStruct* ust)
{
    UInt16 id, usys;
    MemHandle dh;
    UnitStruct u;
    UnitStruct* dp;

    AddUnitStruct(ust);

    InitializeUnitStruct(&u);
    u.usys=ust->unit;
    u.unit=1;
    id=DmFindSortPosition(udb, &u, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return;
    id--;
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmGetRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys>=u.usys){
            usys=dp->usys;
            usys++;
            DmWrite(dp, offsetof(UnitStruct, usys), &usys, sizeof(usys));
        }
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        id++;
    }
    AddUnitStruct(&u);
//    if(usys_time>=u.usys) usys_time++;
}

Boolean DeleteUnitSystemStruct(UnitStruct* ust)
{
    Boolean ret=False;
    UInt16 id, usys;
    MemHandle dh;
    UnitStruct u;
    UnitStruct* dp;

    if(!DeleteUnitStruct(ust)) return ret;

    InitializeUnitStruct(&u);
    u.usys=ust->unit;
    u.unit=1;
    id=DmFindSortPosition(udb, &u, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    id--;
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmQueryRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys==u.usys){
            MemHandleUnlock(dh);
            DmDeleteRecord(udb, id);
            DmMoveRecord(udb, id, DmNumRecords(udb));
        }else if(dp->usys>u.usys){
            MemHandleUnlock(dh);
            break;
        }
//        id++;
    }
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmGetRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys>u.usys){
            usys=dp->usys;
            usys--;
            DmWrite(dp, offsetof(UnitStruct, usys), &usys, sizeof(usys));
        }
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        id++;
    }
    if(GetUnitSystemNum(1)==0){
        UnitStruct u;

        InitializeUnitStruct(&u);
        u.usys=1;  u.unit=1;
        AddUnitStruct(&u);
    }
/*    if(usys_time==u.usys){
        SeekHMS();
    }else if(usys_time>u.usys){
        usys_time--;
    }*/
    return True;
}

Boolean DownUnitSystemStruct(UnitStruct* ust)
{
    Boolean ret=False;
    UInt16 id, usys;
    MemHandle dh;
    UnitStruct u;
    UnitStruct* dp;

    if(!DownUnitStruct(ust)) return ret;

    InitializeUnitStruct(&u);
    u.usys=ust->unit;
    u.unit=1;
    id=DmFindSortPosition(udb, &u, 0, (DmComparF*)CompareUnit, 0);
    if(id==0) return ret;
    id--;
    while(id<DmNumRecordsInCategory(udb, dmAllCategories)){
        dh=DmGetRecord(udb, id);
        dp=(UnitStruct*)MemHandleLock(dh);
        if(dp->usys==u.usys){
            usys=dp->usys;
            usys++;
            DmWrite(dp, offsetof(UnitStruct, usys), &usys, sizeof(usys));
        }else if(dp->usys==u.usys+1){
            usys=dp->usys;
            usys--;
            DmWrite(dp, offsetof(UnitStruct, usys), &usys, sizeof(usys));
        }else{
            MemHandleUnlock(dh);
            DmReleaseRecord(udb, id, True);
            break;
        }
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
        id++;
    }
    DmQuickSort(udb, (DmComparF*)CompareUnit, 0);
/*    if(usys_time==u.usys){
        usys_time++;
    }else if(usys_time==u.usys+1){
        usys_time--;
    }*/
    return True;
}


/* initialize */

void InitializeUnitDB(void)
{
    int i;
    UInt16 id;
    MemHandle dh;
    MemPtr dp;

#ifdef DB_DEFAULT
    UnitStruct ust[] =
        {
        /* unit systems */
            {0, 1, 1.0,    "", "none"},
            {0, 2, 1.0,    "", "length"},
            {0, 3, 1.0,    "", "weight"},
            {0, 4, 1.0,    "", "time"},
            {0, 5, 1.0,    "", "area"},
            {0, 6, 1.0,    "", "currency"},
            {0, 7, 1.0,    "", "speed"},
            {0, 8, 1.0,    "", "angle"},
        /* none */
            {1, 1, 1.0,    "", "" },
        /* length */
            {2, 1, 1.0,    "m",  "m"},
            {2, 2, 1000,   "km", "km"},
            {2, 3, 0.01,   "cm", "cm"},
        /* weight */
            {3, 1, 1.0,    "g",  "g" },
            {3, 2, 1000,   "kg", "kg"},
            {3, 3, 0.001,  "mg", "mg"},
        /* time */
            {4, 1, 1.0,   "",  "HMS"},
            {4, 2, 1.0,   "sec",  "sec"},
            {4, 3, 60,    "min",  "min"},
            {4, 4, 3600,  "hour", "hour"},
        /* area */
            {5, 1, 1.0,    "m2", "m2"},
            {5, 2, 1e+6,   "km2", "km2"},
            {5, 3, 1e-4,   "cm2", "cm2"},
        /* currency */
            {6, 1, 1.0,    "yen", "yen"},
            {6, 2, 116.0,  "US$", "US$"},
        /* speed */
            {7, 1, 1.0,    "km/h", "km/h"},
            {7, 2, 1.609,  "mph",  "mile/h"},
        /* angle */
            {8, 1, 1.0,       "deg",  "degree"},
            {8, 2, 180.0/PI,  "rad",  "radian"},
            {8, 3, 180.0,     "pi",   "pi"}
        };
    int nust=sizeof(ust)/sizeof(*ust);
#endif /* DB_DEFAULT */
#ifdef DB_ENG_JPN
    UnitStruct ust[] =
        {
        /* unit systems */
            {0, 1, 1.0,    "", "none"},
            {0, 2, 1.0,    "", "length"},
            {0, 3, 1.0,    "", "weight"},
            {0, 4, 1.0,    "", "time"},
            {0, 5, 1.0,    "", "area"},
            {0, 6, 1.0,    "", "currency"},
            {0, 7, 1.0,    "", "speed"},
            {0, 8, 1.0,    "", "angle"},
        /* none */
            {1, 1, 1.0,    "", "" },
        /* length */
            {2, 1, 1.0,    "m",  "m"},
            {2, 2, 1000,   "km", "km"},
            {2, 3, 0.01,   "cm", "cm"},
            {2, 4, 0.001,  "mm", "mm"},
            {2, 5, 0.9144, "yd", "yard"},
            {2, 6, 0.3048, "ft", "feet"},
            {2, 7, 0.0254, "in", "inch"},
            {2, 8, 1609.3, "mi", "mile"},
            {2, 9, 0.30303, "", ""},
            {2, 10, 0.030303, "", ""},
            {2, 11, 1.8182, "", ""},
            {2, 12, 3927.3, "", ""},
            {2, 13, 0.0003515, "pt", "pt"},
            {2, 14, 9.46055e+15, "N", "N"},
            {2, 15, 3.0857e+16,  "pc",   "pc"},
        /* weight */
            {3, 1, 1.0,    "g",  "g" },
            {3, 2, 1000,   "kg", "kg"},
            {3, 3, 0.001,  "mg", "mg"},
            {3, 4, 453.6,  "lb", "pound"},
            {3, 5, 28.35,  "oz", "ounce"},
        /* time */
            {4, 1, 1.0,   "",  "HMS"},
            {4, 2, 1.0,   "sec",  "sec"},
            {4, 3, 60,    "min",  "min"},
            {4, 4, 3600,  "hour", "hour"},
            {4, 5, 86400, "day",  "day"},
        /* area */
            {5, 1, 1.0,    "m2", "m2"},
            {5, 2, 1e+6,   "km2", "km2"},
            {5, 3, 1e-4,   "cm2", "cm2"},
            {5, 4, 100,    "a", "A["},
            {5, 5, 10000,  "ha", "wN^["},
            {5, 6, 0.092903,   "ft2", "feet2"},
            {5, 7, 2.59e+6,    "mi2", "mile2"},
            {5, 8, 4046.9, "ac", "G[J["},
            {5, 9, 3.3058, "", ""},
        /* currency */
            {6, 1, 1.0,    "yen", "yen"},
            {6, 2, 116.0,  "US$", "US$"},
            {6, 3, 117.1,  "Eur", "EURO"},
            {6, 4, 182.9,  "",  ""},
            {6, 5, 64.65,  "AU$", "AU$"},
            {6, 6, 0.10,   "won", "won"},
        /* speed */
            {7, 1, 1.0,    "km/h", "km/h"},
            {7, 2, 1.609,  "mph",  "mile/h"},
            {7, 3, 1187.9, "mach", "mach"},
        /* angle */
            {8, 1, 1.0,       "deg",  "degree"},
            {8, 2, 180.0/PI,  "rad",  "radian"},
            {8, 3, 180.0,     "pi",   "pi"}
        };
    int nust=sizeof(ust)/sizeof(*ust);
#endif /* DB_ENG_JPN */
#ifdef DB_ENG
    UnitStruct ust[] =
        {
        /* unit systems */
            {0, 1, 1.0,    "", "none"},
            {0, 2, 1.0,    "", "length"},
            {0, 3, 1.0,    "", "weight"},
            {0, 4, 1.0,    "", "time"},
            {0, 5, 1.0,    "", "area"},
            {0, 6, 1.0,    "", "currency"},
            {0, 7, 1.0,    "", "speed"},
            {0, 8, 1.0,    "", "angle"},
        /* none */
            {1, 1, 1.0,    "", "" },
        /* length */
            {2, 1, 1.0,    "m",  "m"},
            {2, 2, 1000,   "km", "km"},
            {2, 3, 0.01,   "cm", "cm"},
            {2, 4, 0.001,  "mm", "mm"},
            {2, 5, 0.9144, "yd", "yard"},
            {2, 6, 0.3048, "ft", "feet"},
            {2, 7, 0.0254, "in", "inch"},
            {2, 8, 1609.3, "mile", "mile"},
            {2, 9, 0.0003515, "pt", "pt"},
            {2, 10, 9.46055e+15, "LY", "light year"},
            {2, 11, 3.0857e+16,  "pc",   "pc"},
        /* weight */
            {3, 1, 1.0,    "g",  "g" },
            {3, 2, 1000,   "kg", "kg"},
            {3, 3, 0.001,  "mg", "mg"},
            {3, 4, 453.6,  "lb", "pound"},
            {3, 5, 28.35,  "oz", "ounce"},
        /* time */
            {4, 1, 1.0,   "",     "HMS"},
            {4, 2, 1.0,   "sec",  "sec"},
            {4, 3, 60,    "min",  "min"},
            {4, 4, 3600,  "hour", "hour"},
            {4, 5, 86400, "day",  "day"},
        /* area */
            {5, 1, 1.0,    "m2", "m2"},
            {5, 2, 1e+6,   "km2", "km2"},
            {5, 3, 1e-4,   "cm2", "cm2"},
            {5, 4, 100,    "a", "are"},
            {5, 5, 10000,  "ha", "hectare"},
            {5, 6, 0.092903,   "ft2", "feet2"},
            {5, 7, 2.59e+6,    "mi2", "mile2"},
            {5, 8, 4046.9, "ac", "acre"},
        /* currency */
            {6, 1, 1.0,    "yen", "yen"},
            {6, 2, 116.0,  "US$", "US$"},
            {6, 3, 117.1,  "Eur", "EURO"},
            {6, 4, 182.9,  "\xA3",  "UK\xA3"},
            {6, 5, 64.65,  "AU$", "AU$"},
            {6, 6, 0.10,   "won", "won"},
        /* speed */
            {7, 1, 1.0,    "km/h", "km/h"},
            {7, 2, 1.609,  "mph",  "mile/h"},
            {7, 3, 1187.9, "mach", "mach"},
        /* angle */
            {8, 1, 1.0,       "deg",  "degree"},
            {8, 2, 180.0/PI,  "rad",  "radian"},
            {8, 3, 180.0,     "pi",   "pi"}
        };
    int nust=sizeof(ust)/sizeof(*ust);
#endif /* DB_ENG */
#ifdef DB_JPN
    UnitStruct ust[] =
        {
        /* unit systems */
            {0, 1, 1.0,    "", "none"},
            {0, 2, 1.0,    "", ""},
            {0, 3, 1.0,    "", "d"},
            {0, 4, 1.0,    "", ""},
            {0, 5, 1.0,    "", "ʐ"},
            {0, 6, 1.0,    "", "ʉ"},
            {0, 7, 1.0,    "", ""},
            {0, 8, 1.0,    "", "px"},
        /* none */
            {1, 1, 1.0,    "", "" },
        /* length */
            {2, 1, 1.0,    "m",  "m"},
            {2, 2, 1000,   "km", "km"},
            {2, 3, 0.01,   "cm", "cm"},
            {2, 4, 0.001,  "mm", "mm"},
            {2, 5, 0.9144, "yd", "[h"},
            {2, 6, 0.3048, "ft", "tB[g"},
            {2, 7, 0.0254, "in", "C`"},
            {2, 8, 1609.3, "mile", "}C"},
            {2, 9, 0.30303, "", ""},
            {2, 10, 0.030303, "", ""},
            {2, 11, 1.8182, "", ""},
            {2, 12, 3927.3, "", ""},
            {2, 13, 0.0003515, "pt", "|Cg"},
            {2, 14, 9.46055e+15, "N", "N"},
            {2, 15, 3.0857e+16,  "pc",   "p[ZN"},
        /* weight */
            {3, 1, 1.0,    "g",  "g" },
            {3, 2, 1000,   "kg", "kg"},
            {3, 3, 0.001,  "mg", "mg"},
            {3, 4, 453.6,  "lb", "|h"},
            {3, 5, 28.35,  "oz", "IX"},
        /* time */
            {4, 1, 1.0,   "",    "HMS"},
            {4, 2, 1.0,   "b",  "b"},
            {4, 3, 60,    "",  ""},
            {4, 4, 3600,  "",  ""},
            {4, 5, 86400, "",  ""},
        /* area */
            {5, 1, 1.0,    "m2", "m2"},
            {5, 2, 1e+6,   "km2", "km2"},
            {5, 3, 1e-4,   "cm2", "cm2"},
            {5, 4, 100,    "a", "A["},
            {5, 5, 10000,  "ha", "wN^["},
            {5, 6, 0.092903,   "ft2", "tB[g"},
            {5, 7, 2.59e+6,    "mi2", "}C"},
            {5, 8, 4046.9, "ac", "G[J["},
            {5, 9, 3.3058, "", ""},
        /* currency */
            {6, 1, 1.0,    "~",  "~"},
            {6, 2, 116.0,  "US$", "hiāj"},
            {6, 3, 117.1,  "Eur", "["},
            {6, 4, 182.9,  "",  "|h"},
            {6, 5, 64.65,  "AU$", "hij"},
            {6, 6, 0.10,   "won", "EH"},
        /* speed */
            {7, 1, 1.0,    "km/h", "L"},
            {7, 2, 1.609,  "mph",  "}C"},
            {7, 3, 1187.9, "mach", "}bn"},
        /* angle */
            {8, 1, 1.0,       "",   "x"},
            {8, 2, 180.0/PI,  "rad",  "WA"},
            {8, 3, 180.0,     "",   ""}
        };
    int nust=sizeof(ust)/sizeof(*ust);
#endif /* DB_JPN */

    for(i=0; i<nust; i++){
        id=dmMaxRecordIndex;
        dh=DmNewRecord(udb, &id, sizeof(ust[i]));
        dp=(UnitStruct*)MemHandleLock(dh);
        DmWrite(dp, 0, ust+i, sizeof(ust[i]));
        MemHandleUnlock(dh);
        DmReleaseRecord(udb, id, True);
    }
}


Err CreateUnitDB(void)
{
    Err err=0;

    err=DmCreateDatabase(0, UDB_NAME, CREATOR_ID, UDB_TYPE, False);
    if(err) return err;
    udb=DmOpenDatabaseByTypeCreator(UDB_TYPE, CREATOR_ID, dmModeReadWrite);
    if(!udb) return DmGetLastErr();
    InitializeUnitDB();
    DmCloseDatabase(udb);

    return err;
}

Err OpenUnitDB(Boolean md)
{
    Err err=0;
    UInt16 mode;

    if(udb) CloseUnitDB();
    if(md){
        mode=dmModeReadWrite;
    }else{
        mode=dmModeReadOnly;
    }
    udb=DmOpenDatabaseByTypeCreator(UDB_TYPE, CREATOR_ID, mode);
    if(!udb){
        if((err=CreateUnitDB())) return err;
        udb=DmOpenDatabaseByTypeCreator(UDB_TYPE, CREATOR_ID, mode);
        if(!udb) return DmGetLastErr();
    }
    SeekHMS();
    return err;
}

void CloseUnitDB(void)
{
    if(!udb) return;
    DmCloseDatabase(udb);
    udb=0;
}


